/**
 * Copyright 2008 Peach Jean Solutions
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.xj4.util;

import java.util.Collection;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.net.URL;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.commons.lang.UnhandledException;
import org.apache.commons.collections15.Predicate;

/**
 * This is a service loader class, based on the concept of jar services.  All resources that match the name
 * <code>META-INF/services/${serviceClass}</code> will be loaded, and each line will be treated as a class name.
 */
public class ServiceLoader<T> implements Iterable<T> {

  private static final Log logger = LogFactory.getLog(ServiceLoader.class);

  private static final String SERVICE_FILE_PREFIX = "META-INF/services/";

  private Class<T> serviceClass;

  private Collection<T> providers;
  private Object[] constructorArgs;
  private Class[] constructorTypes;

  private List<T> prepend = new ArrayList<T>();
  private List<T> append = new ArrayList<T>();

  private static Class[] calculateTypes(Object[] constructorArgs) {
    Class[] constructorTypes = new Class[constructorArgs.length];
    for(int i = 0; i < constructorArgs.length; i++) {
      constructorTypes[i] = constructorArgs[i] == null ? null : constructorArgs[i].getClass();
    }
    return constructorTypes;
  }

  public ServiceLoader(Class<T> serviceClass, Object ... constructorArgs) {
    this(serviceClass, calculateTypes(constructorArgs), constructorArgs);
  }

  public ServiceLoader(Class<T> serviceClass, Class[] constructorTypes, Object ... constructorArgs) {
    this.serviceClass = serviceClass;
    this.constructorArgs = constructorArgs;
    this.constructorTypes = constructorTypes;
  }

  public void reset() {
    providers = null;
  }

  public void fullReset() {
    prepend.clear();
    append.clear();
  }

  public Iterator<T> iterator() {
    try {
      checkAndReload();
    } catch (IOException e) {
      throw new UnhandledException("Could not load service providers...", e);
    }
    return providers.iterator();
  }

  public void prepend(T ... values) {
    for(T value: values) {
      prepend.add(value);
    }
  }

  public void append(T ... values) {
    for(T value: values) {
      append.add(value);
    }
  }

  private void checkAndReload() throws IOException {
    if(providers == null) {
      reload();
    }
  }

  private void reload() throws IOException {
    Collection<T> providers = new ArrayList<T>();

    providers.addAll(prepend);

    Enumeration<URL> resources = this.getClass().getClassLoader().getResources(SERVICE_FILE_PREFIX + serviceClass.getName());
    while(resources.hasMoreElements()) {
      URL resource = resources.nextElement();
      BufferedReader reader = null;
      try {
        reader = new BufferedReader(new InputStreamReader(resource.openStream()));
        String line = null;
        while((line = reader.readLine()) != null) {
          try {
            Class<? extends T> providerClass = (Class<? extends T>)Class.forName(line.trim());
            providers.add(providerClass.getConstructor(constructorTypes).newInstance(constructorArgs));
          } catch (Exception e) {
            logger.error("Could not handle line: " + line);
          }
        }
      } catch (Exception e) {
        logger.error("Could not handle service resource: " + resource.toString());
      } finally {
        if(reader != null) {
          try {
            reader.close();
          } catch (IOException e) {
            // swallow quietly
          }
        }
      }
    }

    providers.addAll(append);
    
    this.providers = providers;
  }

  public T selectFirst(Predicate<T> selector) {
    for(T t: this) {
      if(selector.evaluate(t)) {
        return t;
      }
    }
    return null;
  }

  public T selectFirstRequired(Predicate<T> selector) {
    T t = selectFirst(selector);
    if(t == null) {
      throw new IllegalStateException("No matching provider found.");
    }
    return t;
  }

  public List<T> selectAll(Predicate<T> selector) {
    List<T> all = new ArrayList<T>();
    for(T t: this) {
      if(selector.evaluate(t)) {
        all.add(t);
      }
    }
    return all;
  }
  
}
